/*
 * Time handling and initialization details
 * 
 * Copyright (C) 1991, 1992, 1995  Linus Torvalds
 * Copyright (C) 1996, 1997, 1998  Ralf Baechle
 * Copyright (C) 2000-2005 TiVo Inc. All Rights Reserved.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/param.h>
#include <linux/interrupt.h>
#ifdef CONFIG_KERNPROF
#include <linux/kernprof.h>
#endif
#include <linux/timex.h>
#include <linux/tivo-monotime.h>

#include <asm/mipsregs.h>
#include <asm/bcm7315.h>

extern int tivo_setup_irq(int, struct irqaction*);
extern volatile unsigned long wall_jiffies;
extern rwlock_t xtime_lock;

#define USECS_PER_JIFFY (1000000/HZ)
#define COUNT_PER_JIFFY (((unsigned int)27000000)/HZ)

// this must always be called with xtime_lock held and IRQs disabled
static inline unsigned long do_gettimeoffset(void)
{
	unsigned int cur_count;
	unsigned long usecs;

	cur_count = *BCM_TIMER0_STAT & 0x3fffffff;
	// usecs = cur_count * USEC_PER_JIFFY / COUNT_PER_JIFFY
	//       = cur_count * 1000000 / 27000000
	//       = cur_count * ((1 << 32) / 27) >> 32
	//       = (cur_count * 159072863) >> 32
	__asm__(" multu %1, %2;"
	        " mfhi  %0;"
	        : "=r" (usecs)
	        : "r" (cur_count), "r" (159072863)
	        : "lo", "hi"
	);
        barrier();
        // Detect and handle just-wrapped case (but be careful not
        // to false-trigger if interrupt went active after we read
        // BCM_TIMER0_STAT):
        if ( (*BCM_TINT & 0x01) && (cur_count < 500) ) {
		usecs += USECS_PER_JIFFY;
	}
	return usecs;
}

/*
 * This version of gettimeofday has 1 microsecond precision.
 */
void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;
	unsigned long usec, sec, lost;

	read_lock_irqsave(&xtime_lock, flags);
	sec = xtime.tv_sec;
	usec = xtime.tv_usec + do_gettimeoffset();
	/*
	 * xtime is atomically updated in timer_bh. jiffies > wall_jiffies
	 * if the timer bottom half hasnt executed yet.
	 */
	lost = jiffies - wall_jiffies;
	read_unlock_irqrestore(&xtime_lock, flags);

	if (lost)
		usec += lost * USECS_PER_JIFFY;

	while (usec >= 1000000) {
		usec -= 1000000;
		sec++;
	}

	tv->tv_sec = sec;
	tv->tv_usec = usec;
}

void do_settimeofday(struct timeval *tv)
{
	write_lock_irq(&xtime_lock);
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= do_gettimeoffset();

	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		/* stop active adjtime() */
	time_status |= STA_UNSYNC;
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	write_unlock_irq(&xtime_lock);
}

#if defined(CONFIG_KERNPROF)
int prof_multiplier[NR_CPUS] = { [0 ... NR_CPUS - 1] = 1 };
int prof_counter[NR_CPUS] = { [0 ... NR_CPUS - 1] = 1 };

/*
 * Change the frequency of the profiling timer.  The multiplier is specified
 * by an appropriate ioctl() on /dev/kernprof.
 */
int setup_profiling_timer(unsigned int multiplier)
{
       int i;

       /*
        * Sanity check.
        
       if ( (!multiplier) || (NSEC_PER_CYCLE/multiplier < 50))
               return -EINVAL;
	*/
       
       /*
        * Set the new multiplier for each CPU. CPUs don't start using the
        * new values until the next timer interrupt in which they do process
        * accounting. At that time they also adjust their APIC timers
        * accordingly.
        */
       for (i = 0; i < NR_CPUS; ++i)
               prof_multiplier[i] = multiplier;

       return 0;
}
#endif




/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick
 */
static void timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
#if defined(CONFIG_KERNPROF)
	if (prof_timer_hook)
		prof_timer_hook(regs);

	if (--prof_counter[0] > 0) {
		write_unlock(&xtime_lock);
		return;
	}

       prof_counter[0] = prof_multiplier[0];
#endif
	*BCM_TINT = 0x010f;
	do_timer(regs);
}

static struct irqaction timerint = { timer_interrupt, SA_INTERRUPT, 0, "timer", NULL, NULL };

void __init time_init(void)
{
	unsigned long count1, count2, old_tcpu1, old_tcpu2;
	unsigned long flags;
	int bail_count;

	// Init to Jan 02-Jan-1970 00:00, so don't get negative time,
	// even after adjusting to local time
	xtime.tv_sec = 24*3600;
	xtime.tv_usec = 0;

	// calibrate the monotonic time counter
	old_tcpu1 = old_tcpu2 = 0;
	bail_count = 500;
	*BCM_TIMER0_CNTL = 0;
	barrier();
	*BCM_TIMER0_CNTL = 0x8003ffff;
	save_and_cli(flags);
	do {
		u32 comp;

		old_tcpu2= old_tcpu1;
		old_tcpu1 = tivo_counts_per_256usec;
		comp = *BCM_TIMER0_STAT;
		while (*BCM_TIMER0_STAT == comp) barrier();
		count1 = read_32bit_cp0_register(CP0_COUNT);
                comp = *BCM_TIMER0_STAT;
                while (((*BCM_TIMER0_STAT - comp) & 0x3ffff) < 512 * 27) barrier();
		count2 = read_32bit_cp0_register(CP0_COUNT);
		tivo_counts_per_256usec = ((count2 - count1) + 1) / 2;
	} while (((tivo_counts_per_256usec != old_tcpu1) || 
		  (old_tcpu1 != old_tcpu2)) && (--bail_count > 0));
	restore_flags(flags);

	*BCM_TIMER0_CNTL = 0;
	printk("Monotonic time calibrated: %ld.%02ld counts per usec\n",
	       tivo_counts_per_256usec / 256,
	       ((tivo_counts_per_256usec & 0xff) * 100) / 256);
	
	*BCM_TIMER0_CNTL = 0x80000000 | (COUNT_PER_JIFFY - 1);
	*BCM_TIMER1_CNTL = 0;
	*BCM_TIMER2_CNTL = 0;
	*BCM_TIMER3_CNTL = 0;
	*BCM_TINT = 0x010f;
	tivo_setup_irq(BCM_LINUX_TIMR_IRQ, &timerint);
}

int __init get_counter_irq(void)
{
	return BCM_LINUX_SYSTIMER_IRQ;
}
