/* $Id: ide-std.c,v 1.4 1999/06/11 14:29:45 ralf Exp $
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * IDE routines for typical pc-like standard configurations.
 *
 * Copyright 2008 TiVo Inc. All Rights Reserved. by Ralf Baechle
 */
#include <linux/sched.h>
#include <linux/ide.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>
#include <asm/ptrace.h>
#include <asm/hdreg.h>
#include <asm/io.h>

#include <linux/tivoconfig.h>
#include <linux/delay.h>
#include <asm/bcm7401.h>

static void (*tivo_cmd_hook)(ide_drive_t *, unsigned char *);
static unsigned int tivo_size_priv(int dma_chunks, int block_chunks);
static void tivo_init_drvpriv(struct idetivo_request *trq, int reading);
static int tivo_add_dmachunk(struct idetivo_request *trq, unsigned long k_addr,
                             unsigned long u_addr, int size, int reading);
static void tivo_add_blkchunk(struct idetivo_request *trq, int chunks, int sectors);
static void tivo_end_blkchunk(struct idetivo_request *trq, int chunks, int sectors);
static void tivo_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *)); 

static struct ll_dma_ops tivo_sgops = {
    size_priv   :   tivo_size_priv,
    init_drvpriv:   tivo_init_drvpriv,
    add_dmachunk:   tivo_add_dmachunk,
    add_blkchunk:   tivo_add_blkchunk,
    end_blkchunk:   tivo_end_blkchunk,
    set_cmd_hook:   tivo_set_cmd_hook,
};

struct tivo_sgent {
    u32 address;
    u32 count;
};

struct desc_chunk {
    unsigned int sglist;   //bus address of scatter list
};

struct tivo_ide_priv {
    struct desc_chunk *first_chunk;
    struct desc_chunk *cur_chunk;
    struct tivo_sgent *first_sgent;
    struct tivo_sgent *cur_sgent;
    int big_flush;
    int last_size;
    u32 bus_addr_end;
};

/*
 These functions are used to preprocess TiVo I/O requests

 Report back how much room needed to allocate along with the request
 for use by the DMA engine, based on number of scatter list entries,
 and number of distinct scatter lists.  Each separate IDE transfer
 (of max MAX_NRSECTORS sectors) requires a separate scatter list.
*/
static unsigned int tivo_size_priv(int dma_chunks, int block_chunks)
{
   unsigned int total_chunk_size = dma_chunks * sizeof(struct tivo_sgent);

   unsigned int size = sizeof(struct tivo_ide_priv) +
               block_chunks * sizeof(struct desc_chunk) +
               (block_chunks - 1) * 8 + // worst case PRD table alignment
               L1_CACHE_BYTES - 4 +  // worst case cache line alignment
               total_chunk_size;
    return size;
}

/*
  put the chunk list right after the priv struct
  put the scatter list after the chunk list (but cache line align it)
*/
static void tivo_init_drvpriv(struct idetivo_request *trq, int reading)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    unsigned long tmp;

    priv->first_chunk = (struct desc_chunk *)((unsigned long)priv +
                                              sizeof(struct tivo_ide_priv));

    tmp = (unsigned long)(priv->first_chunk + trq->num_chunks);
    priv->first_sgent = (struct tivo_sgent *)L1_CACHE_ALIGN(tmp);

    priv->cur_sgent = priv->first_sgent;
    priv->cur_chunk = priv->first_chunk;
    priv->cur_chunk->sglist = virt_to_phys(priv->cur_sgent);

    // this is a bit of platform specificness, but there's no generic way
    // to do it: if the request totals more than twice the dcache size,
    // mark to do one big flush, instead of flushing each individual chunk
    //
    priv->big_flush = (trq->req->nr_sectors > 16 * 1024 / SECTOR_BYTES);

    priv->bus_addr_end = 1; // an impossible addr
}


/*
    Add a chunk to the DMA scatter list and do any necessary cache flushing
    Both kernel and user virtual addresses are passed in, in case the user
    address is needed for flushing purposes (to avoid virtual aliasing
    problem on CPUs with virually indexed caches).  The user address is
    guaranteed not to page fault and size is guaranteed to be non-zero.
*/
static int tivo_add_dmachunk(struct idetivo_request *trq, unsigned long k_addr,
                             unsigned long u_addr, int size, int reading)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    struct tivo_sgent *sgent = priv->cur_sgent;
    u32 bus_addr;
    int total_size;

    if ((k_addr | size) & 3) {
        printk(KERN_WARNING "tivo_ide: DMA not word aligned\n");
        return -EINVAL;
    }

    if (!priv->big_flush) {
        if (reading) {
            dma_cache_inv(u_addr, size);
        } else {
            dma_cache_wback_inv(u_addr, size);
        }
    }

    bus_addr = (u32)virt_to_bus((void *)k_addr);
    // Consolidate adjacent chunks
    //
    // Note that total size will never exceed 128K, because we start a
    // new scatter list at least every 128K
    //
    total_size = size;
    u32 cur_len = total_size;


    while (cur_len) {
        u32 xcount, bcount = 0x10000 - (bus_addr & 0xffff);
        
        if (bcount > cur_len)
            bcount = cur_len;

        xcount = bcount & 0xffff;

#ifdef CONFIG_TIVO_CPU_TO_BE32
        sgent->address = cpu_to_be32(bus_addr);
        sgent->count = cpu_to_be32(xcount);
#else
        sgent->address = cpu_to_le32(bus_addr);
        sgent->count = cpu_to_le32(xcount);
#endif /* CONFIG_TIVO_CPU_TO_BE32 */
        bus_addr += bcount;
        cur_len -= bcount;

        priv->cur_sgent = sgent + 1;
        sgent = priv->cur_sgent;
    }
    priv->last_size = total_size;
    priv->bus_addr_end = bus_addr + size;
    return 0;
}
// Finish up a block of DMA chunks that makes up a single IDE transfer,
// and start a new one
//
static void tivo_add_blkchunk(struct idetivo_request *trq, int chunks, int sectors)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    struct desc_chunk *desc = priv->cur_chunk++;

#ifdef CONFIG_TIVO_CPU_TO_BE32
    (priv->cur_sgent - 1)->count |= cpu_to_be32(0x80000000);
#else
    (priv->cur_sgent - 1)->count |= cpu_to_le32(0x80000000);
#endif
    // Align start of PRD table to 16-bytes
    priv->cur_sgent = (struct tivo_sgent *)
        (((unsigned long)priv->cur_sgent + 0xf) & ~0xf);
    (desc + 1)->sglist = virt_to_phys(priv->cur_sgent);

    priv->bus_addr_end = 1; // an impossible addr
}

// Finish up the last block of DMA chunks, and do any final prep on the
// data going to the DMA engine
//
static void tivo_end_blkchunk(struct idetivo_request *trq, int chunks, int sectors)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    unsigned long sg_start = (unsigned long)priv->first_sgent;
    unsigned long sg_end = (unsigned long)priv->cur_sgent;

#ifdef CONFIG_TIVO_CPU_TO_BE32
    (priv->cur_sgent - 1)->count |= cpu_to_be32(0x80000000);
#else
    (priv->cur_sgent - 1)->count |= cpu_to_le32(0x80000000);
#endif
    if (!priv->big_flush) {
        dma_cache_wback_inv(sg_start, sg_end - sg_start);
    } else {
        flush_dcache();
    }
    priv->cur_chunk = priv->first_chunk;
#ifdef CONFIG_TIVO_IDE_ARC
    if (trq->req->cmd == READ) ARC_start(sg_start, sg_end-sg_start);
#endif
}

static void tivo_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *)) {
    tivo_cmd_hook = handler;
}

ide_startstop_t bcm2_dma_intr(ide_drive_t *drive)
{
    int i, dma_stat;
    byte stat;
    struct request *rq = HWGROUP(drive)->rq;
    struct idetivo_request *trq = rq->tivo;

#ifdef CONFIG_TIVO_IDE_ARC
    ARC_end();
#endif
    dma_stat = HWIF(drive)->dmaproc(ide_dma_end, drive);
    stat = GET_STAT();            /* get drive status */
    if (OK_STAT(stat,DRIVE_READY,drive->bad_wstat|DRQ_STAT)) {
        if (!dma_stat) {
            plog_diskstop(rq);
            if (trq) {
                unsigned int max_nrsectors = trq->max_nrsectors;

                if (rq->nr_sectors > max_nrsectors) {
                    ide_hwif_t *hwif = HWIF(drive);
                    ide_startstop_t startstop;

                    rq->nr_sectors -= max_nrsectors;
                    rq->sector += max_nrsectors;
                    trq->progress += max_nrsectors * SECTOR_BYTES;
                    add_blkdev_randomness(MAJOR(rq->rq_dev));
                    SELECT_DRIVE(hwif, drive);
                    if (ide_wait_stat(&startstop, drive, drive->ready_stat,
                                      BUSY_STAT|DRQ_STAT, WAIT_READY)) {
                        printk("%s: drive not ready for command\n", drive->name);
                        return startstop;
                    }
					//printk("bcm2_dma_intr giving more commands\n");
                    return DRIVER(drive)->do_request(drive, rq, rq->sector
                                                      + trq->partition_start);
                } else {
                    trq->progress += rq->nr_sectors * SECTOR_BYTES;
                ide_end_request(1, HWGROUP(drive));
                trq->status = IDETIVO_DONE;
                return ide_stopped;
                }
            } else {
                for (i = rq->nr_sectors; i > 0;) {
                    i -= rq->current_nr_sectors;
                    ide_end_request(1, HWGROUP(drive));
                }
            }
            return ide_stopped;
        }
        printk("%s: dma_intr: bad DMA status\n", drive->name);
    }
    if (trq) {
        struct tivo_ide_priv *priv = trq->drv_priv;

        // if the request gets retried, we need to reprocess this
        // chunk of the DMA scatterlist
        priv->cur_chunk--;
        trq->status = IDETIVO_ERROR;
        trq->drive_status = stat;
        // XXX: should probably read drive regs to query how far we
        // got in last 256 sector (max) chunk, too...
    }
    return ide_error(drive, "dma_intr", stat);
}

//#ifdef BCM_IDE_DEBUG
static inline void print_sg_list(struct tivo_ide_priv *priv)
{
    struct tivo_sgent *sgent = priv->first_sgent;
    unsigned long sg_end = (unsigned long)priv->cur_sgent;

    while ((unsigned long)sgent != sg_end) {
        unsigned long sg_address = be32_to_cpu(sgent->address);
        unsigned long sg_count = sgent->count;
        if (sg_count & 0x80000000) {
            sg_count &= 0x7FFFFFFF;
            sg_count = be32_to_cpu(sg_count);
        }
        printk("sg_address = 0x%x sg_count = %d\n", sg_address, sg_count);
        sgent++;
    }
}
//#endif /* BCM_IDE_DEBUG */

int bcm2_build_dmatable(ide_drive_t *drive) 
{
    struct request *rq = HWGROUP(drive)->rq;
    
    if (!rq->tivo) {
        return 1;
    }
    struct tivo_ide_priv *priv = rq->tivo->drv_priv;
    struct desc_chunk *chunk = priv->cur_chunk++;
    ide_hwif_t *hwif = HWIF(drive);

    rq->tivo->status = IDETIVO_INPROGRESS;

    unsigned long dma_base        = hwif->dma_base;

	// print_sg_list(priv);
    /* write the PRD table to the controller */
    TIVO_IDE_DMAREG_OUTL(cpu_to_le32((unsigned long)chunk->sglist), dma_base + 4);
    //outl(cpu_to_le32((unsigned long)chunk->sglist), dma_base + 4); 
    barrier();

    return 1;
}

static int tivo_ide_default_irq(ide_ioreg_t base)
{
    return 0;
}

static ide_ioreg_t tivo_ide_default_io_base(int index)
{
    return 0;
}

static void tivo_ide_init_hwif_ports(hw_regs_t *hw, ide_ioreg_t data_port,
                                    ide_ioreg_t ctrl_port, int *irq)
{
    ide_ioreg_t reg = data_port;
    int i;
    TivoConfigValue ide_cap;

    for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
        hw->io_ports[i] = reg;
        reg += 1;
    }
    if (ctrl_port) {
        hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
    } else {
        hw->io_ports[IDE_CONTROL_OFFSET] = hw->io_ports[IDE_DATA_OFFSET] + 0x206;
    }
    if (irq != NULL)
        *irq = 0;
    hw->io_ports[IDE_IRQ_OFFSET] = 0;
#if 1    
    if (GetTivoConfig(kTivoConfigIdeCap, &ide_cap)) {
        ide_hwifs[0].udma_four = ide_hwifs[1].udma_four = 0;
    } else {
        ide_cap &= kTivoConfigIdeCapAta66;
        ide_hwifs[0].udma_four = ide_hwifs[1].udma_four = ide_cap ? 1 : 0;
    }
#endif
    ide_hwifs[0].tivo.version = 
            ide_hwifs[1].tivo.version = IDETIVO_DRIVER_VERSION;
    ide_hwifs[0].tivo.sgops = 
            ide_hwifs[1].tivo.sgops = &tivo_sgops;
}

static int tivo_ide_request_irq(unsigned int irq,
                                void (*handler)(int,void *, struct pt_regs *),
                                unsigned long flags, const char *device,
                                void *dev_id)
{
    return request_irq(irq, handler, flags, device, dev_id);
}            

static void tivo_ide_free_irq(unsigned int irq, void *dev_id)
{
    free_irq(irq, dev_id);
}

static int tivo_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
    return check_region(from, extent);
}

static void tivo_ide_request_region(ide_ioreg_t from, unsigned int extent,
                                    const char *name)
{
    request_region(from, extent, name);
}

static void tivo_ide_release_region(ide_ioreg_t from, unsigned int extent)
{
    release_region(from, extent);
}

struct ide_ops bcm_ide_ops = {
    &tivo_ide_default_irq,
    &tivo_ide_default_io_base,
    &tivo_ide_init_hwif_ports,
    &tivo_ide_request_irq,
    &tivo_ide_free_irq,
    &tivo_ide_check_region,
    &tivo_ide_request_region,
    &tivo_ide_release_region
};

void _tivo_ide_out_byte(u8 data, ide_ioreg_t port)
{
  *((volatile unsigned char *)(mips_io_port_base + port)) = __ioswab32(data);
}

EXPORT_SYMBOL(_tivo_ide_out_byte);

void _tivo_ide_out_word(u16 data, ide_ioreg_t port)
{
  /* 
   * Uh, two spaces per tab?, then hard tabs in bcm_ide_ops?, 
   * What random standard are we writing code to now? 
   * Pick a freaking standard for the file, and stick to it.
   * Learn contextual coding and stop giving folks headaches.
   * I don't know who did it, and don't want to know, when you
   * read this..., fix it, and remove this comment.
   *
   * --snay
   */
  *((volatile unsigned short *)(mips_io_port_base + port)) = __ioswab32(data);
}

EXPORT_SYMBOL(_tivo_ide_out_word);

u8 _tivo_ide_in_byte(ide_ioreg_t port)
{
  return  __ioswab32(*((volatile unsigned char *)(mips_io_port_base + port)));
}

EXPORT_SYMBOL(_tivo_ide_in_byte);

u16 _tivo_ide_in_word(ide_ioreg_t port)
{
  return  __ioswab32(*((volatile unsigned short *)(mips_io_port_base + port)));
}

EXPORT_SYMBOL(_tivo_ide_in_word);
     

/*
 * These are duplicates of code elsewhere, we need it here because it's
 * #ifndef CONFIG_TIVO'ed where TiVo platforms need it to be different,
 * but for PX, we actually need the standard PCI IDE code:
 */

#undef  insw
#define insw(port, addr, count) ide_insw(port, addr, count)
static inline void ide_insw(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(u16 *)addr =  __ioswab32(*(volatile u16 *)(mips_io_port_base + port));
    addr += 2;
  }
}

#undef  outw
#define outw(port, addr, count) ide_outw(port, addr, count)
static inline void ide_outw(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(volatile u16 *)(mips_io_port_base + port) =  __ioswab32(*(u16 *) addr);
    addr += 2;
  }
}

#undef  insl
#define insl(port, addr, count) ide_insl(port, addr, count)
static inline void ide_insl(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(u32 *)addr =  __ioswab32(*(volatile u32 *)(mips_io_port_base + port));
    addr += 4;
  }
}

#undef  outl
#define outl(port, addr, count) ide_outl(port, addr, count)
static inline void ide_outl(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(volatile u32 *)(mips_io_port_base + port) =  __ioswab32(*(u32 *) addr);
    addr += 4;
  }
}


/*
 * This is used for most PIO data transfers *from* the IDE interface
 */
void ide_input_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
    byte io_32bit = drive->io_32bit;

    if (io_32bit) {
        insl(IDE_DATA_REG, buffer, wcount);
    } else {
        insw(IDE_DATA_REG, buffer, wcount<<1);
    }
}

/*
 * This is used for most PIO data transfers *to* the IDE interface
 */
void ide_output_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
    byte io_32bit = drive->io_32bit;

    if (io_32bit) {
        outsl(IDE_DATA_REG, buffer, wcount);
    } else {
        outsw(IDE_DATA_REG, buffer, wcount<<1);
    }
}

/*
 * The following routines are mainly used by the ATAPI drivers.
 *
 * These routines will round up any request for an odd number of bytes,
 * so if an odd bytecount is specified, be sure that there's at least one
 * extra byte allocated for the buffer.
 */
void atapi_input_bytes (ide_drive_t *drive, void *buffer, unsigned int bytecount)
{
    ++bytecount;
    ide_input_data (drive, buffer, bytecount / 4);
    if ((bytecount & 0x03) >= 2)
        insw (IDE_DATA_REG, ((byte *)buffer) + (bytecount & ~0x03), 1);
}

void atapi_output_bytes (ide_drive_t *drive, void *buffer, unsigned int bytecount)
{
    ++bytecount;
    ide_output_data (drive, buffer, bytecount / 4);
    if ((bytecount & 0x03) >= 2)
        outsw (IDE_DATA_REG, ((byte *)buffer) + (bytecount & ~0x03), 1);
}



#ifdef CONFIG_TIVO_SATAPM


#define PSCR0_ADDR 0
#define PSCR1_ADDR 1
#define PSCR2_ADDR 2

/*
 * This code is fundamentally wrong, broken, and awful.  It says that
 * it takes microseconds as a value, and that appears to be a
 * complete lie - the math is wrong.  It also handles round-down
 * badly, and fails to deal with the "we were called just before
 * the jiffy counter ticked over" issue.
 *
 * Needs to be completely rewritten, or (better yet) a standard
 * kernel wait-for-however-menu-time-units routine should be used.
 *
 *                 dplatt@tivo.com 4/1/2008
 */

#define OLD_BAD_MATH

static void sata_pm_sleep(int usec)
{
	unsigned long expire;

#ifdef OLD_BAD_MATH
	expire = (usec * HZ) / 100;
	if (expire == 0)
			expire = 1;

	expire += jiffies;
#else
        /*
         * This is still wrong, units-wise, but at least it deals
         * with round-down and inter-jiffy timing properly and should
         * guarantee that it sleeps for at least as long as the
         * caller requested.
         */
	expire = (usec * HZ) / 100;

	expire += jiffies + 2;
#endif
        do {
		if (0 < (signed long)(jiffies - expire)) {
			return;
		}
	} while (1);
}

typedef struct {

    char            IdePriData;            /* 0x0200 */ /* BAR0 */
    char            IdePriFeatErr;        /* 0x0201 */
    char            IdePriSecCnt;        /* 0x0202 */
    char            IdePriSecNum;        /* 0x0203 */
    char            IdePriCylLow;        /* 0x0204 */
    char            IdePriCylHi;        /* 0x0205 */
    char            IdePriDevHd;        /* 0x0206 */
    char            IdePriCmdStat;        /* 0x0207 */
    unsigned int    unused9[14];        /* 0x0208-0x023C */

    char            unusedbyte1;        /* 0x0240 */
    char            unusedbyte2;        /* 0x0241 */
    char            IdePriDevCStat;        /* 0x0242 */ /* BAR1 + 2 -- alternate status*/
    char            unusedbyte3;        /* 0x0243 */
    unsigned int    unused10[15];        /* 0x0244-0x027C */

    char            IdeSecData;            /* 0x0280 */ /* BAR2 */
    char            IdeSecFeatErr;        /* 0x0281 */
    char            IdeSecSecCnt;        /* 0x0282 */
    char            IdeSecSecNum;        /* 0x0283 */
    char            IdeSecCylLow;        /* 0x0284 */
    char            IdeSecCylHi;        /* 0x0285 */
    char            IdeSecDevHd;        /* 0x0286 */
    char            IdeSecCmdStat;        /* 0x0287 */
    unsigned int    unused11[14];        /* 0x0288-0x02BC */

    char            unusedbyte4;        /* 0x02C0 */
    char            unusedbyte5;        /* 0x02C1 */ /* BAR3 + 2 -- alternate status*/
    char            IdeSecDevCStat;        /* 0x02C2 */
    char            unusedbyte6;        /* 0x02C3 */
    unsigned int    unused12[15];        /* 0x02C4-0x02FC */

    unsigned short  PriBmCmd;            /* 0x0300 IDEPriBMCmd */
    unsigned short  PriBmStat;          /* 0x0302 IDEPriBMStat */
    unsigned int    PriBmDescTableAddr;    /* 0x0304 IDEPriDTPT */
    unsigned short  SecBmCmd;            /* 0x0308 IDESecBMCmd */
    unsigned short  SecBmStat;          /* 0x0302 IDESecBMStat */
    unsigned int    SecBmDescTableAddr;    /* 0x030C IDESecDTPT */

} IdeRegisters;
/* Map data structure to IDE registers
 * SATA_MMIO_BASE  is MMIO base address from SATA MMIO BASE address 
 * from SATA PCI configuration registers 
 */
#define IDEC ((volatile IdeRegisters *) (BCM_PCIIO_BASE + 0x200))


static int current_active_pm = 0x0;
static int second_disk_preset = 0x0;

/* 
 * a) Select port multiplier: Set bit 19-16 of MMIO_BASE_ADDR +0x48 [SCR2] 
 * of Host controller to 0xFh 
 * b) Write offset of the register to read to the primary feature/error 
 * register 
 * c) Write port number (0,1) to the Primary Device Head register 
 * d) Issue Read port multiplier command (0xE4) by writing the command 
 * to the Primary command status register 
 * e) Read value of the register from task file register 
 *
 */

static unsigned int PmReadMulReg(unsigned char portNum, unsigned char addr)
{
    volatile unsigned int *mdio,tmp32,retVal;

    mdio = (unsigned int *)(SATA_MMIO_BASE + 0x48);
    tmp32 = *mdio;
    *mdio = 0x000F0000;

    IDEC->IdePriFeatErr = addr;
    IDEC->IdePriSecCnt = 0;
    IDEC->IdePriSecNum = 0;
    IDEC->IdePriCylLow = 0;
    IDEC->IdePriCylHi  = 0;

    IDEC->IdePriDevHd = portNum;
    IDEC->IdePriCmdStat = 0xE4;

    sata_pm_sleep(10);

    retVal = (((unsigned int)IDEC->IdePriCylHi) & 0x000000FF) << 24;
    retVal |= (((unsigned int)IDEC->IdePriCylLow) & 0x000000FF) << 16;
    retVal |= (((unsigned int)IDEC->IdePriSecNum) & 0x000000FF) << 8;
    retVal |= (((unsigned int)IDEC->IdePriSecCnt) & 0x000000FF);

    *mdio = tmp32; /*restore*/

    return retVal;
}

/* 
 * a) Select port multiplier: Set bit 19-16 of MMIO_BASE_ADDR +0x48 [SCR2] 
 * of Host controller to 0xFh 
 * b) Write offset of the register to read to the primary feature/error 
 * register 
 * c) Write port number (0,1) to the Primary Device Head register 
 * d) Issue write port multiplier command (0xE8) by writing the command to 
 * the Primary command status register 
 *
 */

static void PmWriteMulReg(unsigned char portNum, unsigned char addr, unsigned int data)
{

    volatile unsigned int *mdio,tmp32;

    mdio = (unsigned int *)(SATA_MMIO_BASE + 0x48);

    tmp32 = *mdio;

    *mdio = 0x000F0000;

    IDEC->IdePriFeatErr = addr;
    IDEC->IdePriSecCnt = (unsigned char)(data & 0xFF);
    IDEC->IdePriSecNum = (unsigned char)((data >> 8) & 0xFF);
    IDEC->IdePriCylLow = (unsigned char)((data >> 16) & 0xFF);
    IDEC->IdePriCylHi  = (unsigned char)((data >> 24) & 0xFF);

    IDEC->IdePriDevHd = portNum;
    IDEC->IdePriCmdStat = 0xE8;

    sata_pm_sleep(10);

    *mdio = tmp32; /*restore*/

}

static void DumpGSRRegs()
{
    int addr;
    unsigned int tmp32;

    /*Dump GSR regs*/
    printk("\nGSCR:\n");
    for(addr=0; addr<3; addr++)
    {
        tmp32 = PmReadMulReg(0x0F,addr);
        printk("GSR addr %d val = %x\n",addr,tmp32);
    }
    tmp32 = PmReadMulReg(0x0F,0x40);
    printk("GSR addr 0x40 val = %x\n",tmp32);
}

static void DumpPSCRRegs(unsigned int MaxPorts)
{
    int addr,port;
    unsigned int tmp32;

    /*Dump PSR regs*/
    for(port=0 ; port< MaxPorts; port++)
    for(addr=0; addr<3; addr++)
    {
        tmp32 = PmReadMulReg(port,addr);
        printk("PSR port %d addr %d val = %x\n",port,addr,tmp32);
    }
}

static void sata_pm_clear_errors(byte port)
{
    volatile unsigned int 	*mdio,tmp32;

    mdio = (unsigned int *)(SATA_MEM_BASE + 0x48);
    tmp32 = (unsigned int) port << 16;
    *mdio = tmp32;

    PmWriteMulReg(port, PSCR1_ADDR, 0xFFFFFFFF);  // reset the errors.
}


static unsigned long enable_pm_port(byte port)
{
    unsigned int 			pm_status;
	int			 			cnt;
    volatile unsigned int 	*mdio,tmp32;

    PmWriteMulReg(port, PSCR2_ADDR, 1); /* Reset link*/
    PmWriteMulReg(port, PSCR2_ADDR, 0); /* take device out of reset*/

    mdio = (unsigned int *)(SATA_MEM_BASE + 0x48);
    tmp32 = (unsigned int) port << 16;
    *mdio = tmp32;

	cnt = 0;
	do {
    	// pm_status = *(unsigned int *)(SATA_MEM_BASE + 0x40); // PCSR0
    	sata_pm_sleep(10); /*Wait*/
   		pm_status = PmReadMulReg(port, PSCR0_ADDR);
		cnt++;
		if ((cnt > 15) || (port == 2))
				break;
    	printk("SATA PM Port %d Status (PSR0) = 0x%x\n", port, pm_status);
	} while((pm_status & 0xF) != 0x3);


        sata_pm_sleep(20);  /* add guard time... dplatt */

	return pm_status;
}


static void dump_sata_pm_status(void)
{
	unsigned int			t;

	t = *(unsigned int *)(SATA_MEM_BASE + 0x44);
   	printk("SATA SCR1 = 0x%x\n", t);
	t = *(unsigned int *)(SATA_MEM_BASE + 0x40);
   	printk("SATA SCR0 = 0x%x\n", t);
}


#if 0
void sata_pm_port_assert(void)
{
    volatile unsigned int *mdio,tmp32;

    mdio = (unsigned int *)(SATA_MEM_BASE + 0x48);
    tmp32 = *mdio;

	if (((tmp32 >> 16) & 0xF) != (unsigned int) current_active_pm)
	{
		printk("Port number mismatch between the hw and sw\n");
		show_trace_task_bt(current,1);
		BUG();
	}
}



void sata_pm_port1_assert(void)
{
    volatile unsigned int *mdio,tmp32;

    mdio = (unsigned int *)(SATA_MEM_BASE + 0x48);
    tmp32 = *mdio;

	if (((tmp32 >> 16) & 0xF) != 0x1)
	{
		printk("Port number is not 1\n");
		show_trace_task_bt(current,1);
		BUG();
	}
}


void sata_pm_port0_assert(void)
{
    volatile unsigned int *mdio,tmp32;

    mdio = (unsigned int *)(SATA_MEM_BASE + 0x48);
    tmp32 = *mdio;

	if (((tmp32 >> 16) & 0xF) != 0x0)
	{
		printk("Port number is not 0\n");
		show_trace_task_bt(current,1);
		BUG();
	}
}
#endif

static void ActivatePmPort(unsigned char portNum)
{
    volatile unsigned int *mdio,tmp32;
	int			 			cnt;
    unsigned int 			pm_status;

    mdio = (unsigned int *)(SATA_MEM_BASE + 0x48);
    tmp32 = *mdio;

	if (((tmp32 >> 16) & 0xF) != (unsigned int) current_active_pm)
	{
			BUG();
	}

    // printk("SATA port %d active.\n", portNum);
    tmp32 = (unsigned int)portNum << 16;
    *mdio = tmp32;

#if 0
    tmp32 = *mdio;
	while (((tmp32 >> 16) & 0xF) != (unsigned int) portNum)
	{
		printk("SATA Port Value = 0x%x, portNum = 0x%x\n", tmp32, (unsigned int) portNum);
    	sata_pm_sleep(1); /*Wait*/
    	tmp32 = *mdio;
	}
#endif
}


static void sata_pm_soft_reset(int port)
{
    volatile unsigned int *mdio;
    unsigned int 		tmp32;

    mdio = (unsigned int *)(SATA_MEM_BASE + 0x48);

    *mdio = ((unsigned int) port << 16);

    sata_pm_sleep(20);  /* add guard time... dplatt */

    /*Soft reset*/
    IDEC->IdePriDevCStat = 0x04;              

    sata_pm_sleep(10); /*Wait*/
    IDEC->IdePriDevCStat = 0;                 
    sata_pm_sleep(10);

    printk("Reading SATA PM %d Signature\n", port);
    
    /*Read PM port signature*/
    tmp32  = (((unsigned int)IDEC->IdePriCylHi) & 0x000000FF) << 24;
    tmp32 |= (((unsigned int)IDEC->IdePriCylLow) & 0x000000FF) << 16;
    tmp32 |= (((unsigned int)IDEC->IdePriSecNum) & 0x000000FF) << 8;
    tmp32 |= (((unsigned int)IDEC->IdePriSecCnt) & 0x000000FF);

    printk("SATA PM port %d signature: %x\n", port, tmp32);

    sata_pm_sleep(20);  /* add guard time... dplatt */

}


static unsigned int sata_pm_get_nports(void)
{
	unsigned int	tmp32;

	// Read GSCR2 to find out the number of ports.	
     tmp32 = PmReadMulReg(0x0F, 2);

	 printk("Number of SATA PM ports = %d\n", tmp32);

         sata_pm_sleep(20);  /* add guard time... dplatt */

	 return tmp32;
}



void sata_pm_activate_port(int port)
{
	if (!second_disk_preset)
		   return;	

    if (port != current_active_pm)
    {
        //printk("SATA port %d active.\n", port);
        ActivatePmPort(port);
        current_active_pm = port;
    }
}

EXPORT_SYMBOL(sata_pm_activate_port);


int sata_pm_get_current_active_port(void)
{
	return current_active_pm;
}

EXPORT_SYMBOL(sata_pm_get_current_active_port);

int sata_pm_second_disk_status(void)
{
	return second_disk_preset;
}

EXPORT_SYMBOL(sata_pm_second_disk_status);


/* 
 * a) Select port multiplier: Set bit 19-16 of MMIO_BASE_ADDR +0x48 [SCR2] 
 * of Host controller to 0xFh.
 * b) Reset the device by toggling SRST bit in primary device control/status 
 * register.
 * c) Read PM signature from taskfile register.
 *
 */ 
void EnableAndShowPmSignature()
{
    /* Get signature */
    volatile unsigned int *mdio,tmp32;
    int cnt, i;
    unsigned int data;
    unsigned int pm_status;

    // sata_pm_sleep(500);
    // PmWriteMulReg(0, PSCR2_ADDR, 1); /* Reset link*/
    // sata_pm_sleep(500);
    // PmWriteMulReg(0, PSCR2_ADDR, 0); /* take device out of reset*/

    sata_pm_sleep(500);
    mdio = (unsigned int *)(SATA_MEM_BASE + 0x48);
    tmp32 = *mdio;

    *mdio = 0x000F0000;

    printk("Resetting PM\n");
    
    /*Soft reset*/
    IDEC->IdePriDevCStat = 0x04;              

    sata_pm_sleep(1); /*Wait*/
    IDEC->IdePriDevCStat = 0;                 
    sata_pm_sleep(1);

    printk("Reading PM Signature\n");
    
    /*Read PM signature*/
    tmp32  = (((unsigned int)IDEC->IdePriCylHi) & 0x000000FF) << 24;
    tmp32 |= (((unsigned int)IDEC->IdePriCylLow) & 0x000000FF) << 16;
    tmp32 |= (((unsigned int)IDEC->IdePriSecNum) & 0x000000FF) << 8;
    tmp32 |= (((unsigned int)IDEC->IdePriSecCnt) & 0x000000FF);

    printk("\nPM signature: %x\n\n",tmp32);

	cnt = sata_pm_get_nports(); // Get the number of ports.

	second_disk_preset = 0x0;
	for (i = 0; i < cnt; ++i) {
    	printk("Checking for Drive %d\n", i);
		pm_status = enable_pm_port(i);
		if (((pm_status & 0xF) == 0x3) && i == 1) {
			second_disk_preset = 0x1;
			printk("Found the 2nd Disk\n");
		}
	}
	//printk("Currently second disk is NOT supported.\n");
	//second_disk_preset = 0x0;

	if (!second_disk_preset)
	{
		--cnt;
	}

	for (i = 0; i < cnt-1; ++i) {
		sata_pm_clear_errors(i);
	}

	for (i = 0; i < cnt-1; ++i) {
		sata_pm_soft_reset(i);
	}

    printk("Resetting Drives Complete\n");

    // Set port 0 active
    *mdio = 0;
    current_active_pm = 0;

    sata_pm_sleep(10);  /* add guard time... dplatt */

}
EXPORT_SYMBOL(EnableAndShowPmSignature);


#endif // CONFIG_TIVO_SATAPM

#ifdef CONFIG_TIVO_IDE_ARC

#define BVNF_INTR2_2_R5F_CLEAR              *(volatile unsigned int *)(0xb0103208)
#define BVNF_INTR2_2_R5F_MASK_STATUS        *(volatile unsigned int *)(0xb010320c)
#define BVNF_INTR2_2_R5F_MASK_SET           *(volatile unsigned int *)(0xb0103210)
#define BVNF_INTR2_2_R5F_MASK_CLEAR         *(volatile unsigned int *)(0xb0103214)

#define MEMC_0_ARC_0_CNTRL                      *(volatile unsigned int *)(0xb01061bc)
#define MEMC_0_ARC_0_ADRS_RANGE_LOW             *(volatile unsigned int *)(0xb01061c0)
#define MEMC_0_ARC_0_ADRS_RANGE_HIGH            *(volatile unsigned int *)(0xb01061c4)
#define MEMC_0_ARC_0_READ_RIGHTS_LOW            *(volatile unsigned int *)(0xb01061c8)
#define MEMC_0_ARC_0_READ_RIGHTS_HIGH           *(volatile unsigned int *)(0xb01061cc)
#define MEMC_0_ARC_0_WRITE_RIGHTS_LOW           *(volatile unsigned int *)(0xb01061d0)
#define MEMC_0_ARC_0_WRITE_RIGHTS_HIGH          *(volatile unsigned int *)(0xb01061d4)
#define MEMC_0_ARC_0_VIOLATION_INFO_START_ADDR  *(volatile unsigned int *)(0xb01061d8)
#define MEMC_0_ARC_0_VIOLATION_INFO_END_ADDR    *(volatile unsigned int *)(0xb01061dc)
#define MEMC_0_ARC_0_VIOLATION_INFO_CMD         *(volatile unsigned int *)(0xb01061e0)

#define HIF_RAC 41          // Client for 'Host Interface Read Ahead Cache', aka the cpu

extern void show_registers(struct pt_regs *regs);

static void arc_isr(int irq, void *dev_id, struct pt_regs *regs)
{
    BVNF_INTR2_2_R5F_MASK_SET=~0;   // mask it
    printk("IDE ARC violation at %.8X in %.8X-%.8X, EPC=%lX\n",
            MEMC_0_ARC_0_VIOLATION_INFO_START_ADDR*8,
            MEMC_0_ARC_0_ADRS_RANGE_LOW*8, 
            MEMC_0_ARC_0_ADRS_RANGE_HIGH*8,
            regs->cp0_epc);
    
    // show_registers(regs);
}

void ARC_init(void)
{
    printk("Initializing IDE ARC\n"); 
    MEMC_0_ARC_0_CNTRL=0;           // disable ARC
    BVNF_INTR2_2_R5F_MASK_SET=~0;   // mask it
    MEMC_0_ARC_0_READ_RIGHTS_LOW=(HIF_RAC<32)?~(1<<HIF_RAC):~0;
    MEMC_0_ARC_0_READ_RIGHTS_HIGH=(HIF_RAC<32)?~0:~(1<<(HIF_RAC-32));
    MEMC_0_ARC_0_WRITE_RIGHTS_LOW=(HIF_RAC<32)?~(1<<HIF_RAC):~0;
    MEMC_0_ARC_0_WRITE_RIGHTS_HIGH=(HIF_RAC<32)?~0:~(1<<(HIF_RAC-32));
     
    if (request_irq(BCM_IRQ_BVNF_CPU_INTR_2, arc_isr, SA_INTERRUPT, "IDE ARC", NULL))
        printk(KERN_ERR "WARNING: can't register IDE ARC interrupt, proceeding without it...\n");
    else
        MEMC_0_ARC_0_CNTRL=0x0a;    // enable read and write check
}

void ARC_start(int addr, int len)
{
    if (MEMC_0_ARC_0_CNTRL && (BVNF_INTR2_2_R5F_MASK_STATUS & 1))   // if enabled and masked...
    {
        if ((addr & 0x40000007) || ((addr+len) & 0x40000000) || len < 8)   
        {
            printk(KERN_WARNING "IDE ARC unusable %.8X %X\n", addr, len);
            return;
        }
        
        int physaddr = addr & 0x1fffffff;       // ensure physical
        MEMC_0_ARC_0_ADRS_RANGE_LOW=physaddr/8;
        MEMC_0_ARC_0_ADRS_RANGE_HIGH=(physaddr+len-1)/8;
        BVNF_INTR2_2_R5F_CLEAR=~0;              // kill any pending 
        (void)MEMC_0_ARC_0_VIOLATION_INFO_CMD;  // unlatch violation
        BVNF_INTR2_2_R5F_MASK_CLEAR=1;          // unmask ARC0
    }
}

inline void ARC_end(void)
{
    BVNF_INTR2_2_R5F_MASK_SET=~0;   // mask interrupt 
}   
#endif
